@@ -3,30 +3,20 @@
 #include <io.h>
 #include <string.h>
 #include <process.h>
-
-#ifdef OS2
-#ifdef _MSC_VER
-#define popen(c,m)   _popen(c,m)
-#define pclose(f)    _pclose(f)
-#endif
-#endif
-
-#if defined(WIN32) && defined(_MSC_VER)
-#define popen _popen
-#define pclose _pclose
-#endif
+#include <errno.h>
+#include "popen.h"
+#undef popen
+#undef pclose
+#undef system
 
 #ifndef _NFILE
 #define _NFILE 40
 #endif
 
-static char template[] = "piXXXXXX";
-typedef enum { unopened = 0, reading, writing } pipemode;
-static
-struct {
-    char *command;
-    char *name;
-    pipemode pmode;
+static struct {
+  char *command;
+  char *name;
+  char pmode[4];
 } pipes[_NFILE];
 
 
@@ -38,11 +28,14 @@ struct {
  * Currently, only MSC (running under DOS) and MINGW versions are managed.
  */
 
-#if defined(_MSC_VER) || defined(__MINGW32__)
+#if defined(__MINGW32__)
 
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
+
+#if 0
 static int
-unixshell(p)
-char *p;
+unixshell(char *p)
 {
   static char *shell[] = {"sh", "bash", "csh", "tcsh", "sh32", "sh16", "ksh", NULL};
   char **shellp = shell, *s, *q;
@@ -64,8 +57,7 @@ char *p;
 }
 
 static char *
-slashify(p, s)
-char *p, *s;
+slashify(char *p, char *s)
 {
   if (unixshell(s))
     while (s = strchr(p, '\\')) *s = '/';
@@ -75,8 +67,7 @@ char *p, *s;
 }
 
 static char *
-scriptify(command)
-char *command;
+scriptify(const char *command)
 {
   FILE *fp;
   char *cmd, *name, *s, *p;
@@ -90,25 +81,35 @@ char *command;
     p = s;
   }
   slashify(name, p);
-  if (! unixshell(p)) {
-    realloc(name, strlen(name) + 5);
+  if (! (i = unixshell(p))) {
+    char *p = (char *) realloc(name, strlen(name) + 5);
+    if (p == NULL) {
+	free(cmd);
+	return NULL;
+    }
+    name = p;
     strcat(name, ".bat");
   }
   if (s) sprintf(cmd + strlen(cmd), " %cc ", unixshell(s) ? '-' : '/');
   strcpy(p = cmd + strlen(cmd), name); free(name);
 
-  i = strlen(command);
-  if ( ((fp = fopen(p, "wb")) == NULL) || (fwrite(command, 1, i, fp) < i)
-       || (fputc('\n', fp) == EOF)) {
-    cmd = NULL; 
+  if ((fp = fopen(p, i ? "wb" : "w")) != NULL) {
+    if (! i) fputs("@echo off\n", fp);
+    i = strlen(command);
+    if ((fwrite(command, 1, i, fp) < i) || (fputc('\n', fp) == EOF)) {
+      free(cmd);
+      cmd = NULL;
+    }
+  } else {
+    free(cmd);
+    cmd = NULL;
   }
   if (fp) fclose(fp); 
   return(cmd);
 }
 
 static void
-unlink_and_free(cmd)
-char *cmd;
+unlink_and_free(char *cmd)
 {
   char *s;
 
@@ -118,59 +119,101 @@ char *cmd;
     s = cmd;
   unlink(s); free(cmd);
 }
-
-int
-os_system(cmd)
-char *cmd;
-{
-  char *s;
-  int i;
-
-#if defined(OS2)
-  if (_osmode == OS2_MODE)
-    return(system(cmd));
 #endif
 
-  if ((cmd = scriptify(cmd)) == NULL) return(1);
-  if (s = getenv("SHELL"))
-    i = spawnlp(P_WAIT, s, s, cmd + strlen(s), NULL);
-  else
-    i = system(cmd);
-  unlink_and_free(cmd);
+int
+os_system(const char *cmd)
+{
+  char *cmdexe = getenv("ComSpec");
+  char *cmd1 = quote_cmd(cmd);
+  int i = spawnl(P_WAIT, cmdexe, "cmd.exe", "/c", cmd1, NULL);
+
+  free(cmd1);
   return(i);
 }
-#else
+
+#ifndef PIPES_SIMULATED
+int
+kill (int pid, int sig)
+{
+  HANDLE ph;
+  int retval = 0;
+
+  /* We only support SIGKILL.  */
+  if (sig != SIGKILL)
+    {
+      errno = ENOSYS;
+      return -1;
+    }
+
+  ph = OpenProcess(PROCESS_TERMINATE, FALSE, pid);
+  if (ph)
+    {
+      BOOL status = TerminateProcess(ph, -1);
+
+      if (!status)
+	{
+	  errno = EPERM;
+	  retval = -1;
+	}
+    }
+  else
+    {
+      /* If we cannot open the process, it means we eaither aren't
+	 allowed to (e.g., a process of another user), or such a
+	 process doesn't exist.  */
+      switch (GetLastError ())
+	{
+	  case ERROR_ACCESS_DENIED:
+	    errno = EPERM;
+	    break;
+	  default:
+	    errno = ESRCH;
+	    break;
+	}
+      retval = -1;
+    }
+  CloseHandle (ph);
+  return retval;
+}
+
+char *
+quote_cmd(const char *cmd)
+{
+  char *quoted;
+
+  /* The command will be invoked via cmd.exe, whose behavior wrt
+     quoted commands is to remove the first and the last quote
+     characters, and leave the rest (including any quote characters
+     inside the outer pair) intact.  */
+  quoted = malloc(strlen (cmd) + 2 + 1);
+  sprintf(quoted, "\"%s\"", cmd);
+
+  return quoted;
+}
+#endif
+
+#else  /* !__MINGW32__ */
 #define os_system(cmd) system(cmd)
 #endif
 
 
 FILE *
-os_popen( char *command, char *mode ) {
+os_popen(const char *command, const char *mode )
+{
     FILE *current;
     char *name;
     int cur;
-    pipemode curmode;
-    
-#if defined(OS2) && (_MSC_VER != 510)
-    if (_osmode == OS2_MODE)
-      return(popen(command, mode));
-#endif
+    char curmode[4];
 
-    /*
-    ** decide on mode.
-    */
-    if(strcmp(mode,"r") == 0)
-        curmode = reading;
-    else if(strcmp(mode,"w") == 0)
-        curmode = writing;
-    else
-        return NULL;
+    if (*mode != 'r' && *mode != 'w')
+      return NULL;
+    strncpy(curmode, mode, 3); curmode[3] = '\0';
 
-#if defined(__MINGW32__) || (defined(_MSC_VER) && defined(WIN32))
-    current = popen(command = scriptify(command), mode);
+#if defined(__MINGW32__)
+    current = popen(command, mode);
     cur = fileno(current);
-    pipes[cur].pmode = curmode;
-    pipes[cur].command = command;
+    strcpy(pipes[cur].pmode, curmode);
     return(current);
 #endif
 
@@ -183,66 +226,64 @@ os_popen( char *command, char *mode ) {
     ** If we're reading, just call system to get a file filled with
     ** output.
     */
-    if(curmode == reading) {
+    if (*curmode == 'r') {
         FILE *fp;
         if ((cur = dup(fileno(stdout))) == -1)
 	    return NULL;
-        if ((current = freopen(name, "w", stdout)) == NULL)
+	*curmode = 'w';
+        if ((current = freopen(name, curmode, stdout)) == NULL)
 	    return NULL;
         os_system(command);
         if (dup2(cur, fileno(stdout)) == -1)
 	    return NULL;
 	close(cur);
-        if((current = fopen(name,"r")) == NULL)
+	*curmode = 'r';
+        if ((current = fopen(name, curmode)) == NULL)
             return NULL;
     } else {
-        if((current = fopen(name,"w")) == NULL)
+        if ((current = fopen(name, curmode)) == NULL)
             return NULL;
     }
     cur = fileno(current);
     pipes[cur].name = name;
-    pipes[cur].pmode = curmode;
+    strcpy(pipes[cur].pmode, curmode);
     pipes[cur].command = strdup(command);
     return current;
 }
 
 int
-os_pclose( FILE * current) {
-    int cur = fileno(current),rval;
+os_pclose( FILE * current)
+{
+    int cur = fileno(current);
+    int fd, rval;
 
-#if defined(OS2) && (_MSC_VER != 510)
-    if (_osmode == OS2_MODE)
-      return(pclose(current));
-#endif
-
-#if defined(__MINGW32__) || (defined(_MSC_VER) && defined(WIN32))
+#if defined(__MINGW32__)
     rval = pclose(current);
-    pipes[cur].pmode = unopened;
-    unlink_and_free(pipes[cur].command);
+    *pipes[cur].pmode = '\0';
     return rval;
 #endif
 
     /*
     ** check for an open file.
     */
-    if(pipes[cur].pmode == unopened)
-        return -1;
-    if(pipes[cur].pmode == reading) {
+    switch (*pipes[cur].pmode) {
+    case 'r':
         /*
         ** input pipes are just files we're done with.
         */
         rval = fclose(current);
         unlink(pipes[cur].name);
-    } else {
+	break;
+    case 'w':
         /*
         ** output pipes are temporary files we have
         ** to cram down the throats of programs.
         */
-	int fd;
         fclose(current);
 	rval = -1;
 	if ((fd = dup(fileno(stdin))) != -1) {
-	  if (current = freopen(pipes[cur].name, "r", stdin)) {
+	  char *mode = pipes[cur].pmode; *mode = 'r';
+	  if (current = freopen(pipes[cur].name, mode, stdin)) {
 	    rval = os_system(pipes[cur].command);
 	    fclose(current);
 	    if (dup2(fd, fileno(stdin)) == -1) rval = -1;
@@ -250,11 +291,14 @@ os_pclose( FILE * current) {
 	  }
 	}
         unlink(pipes[cur].name);
+	break;
+    default:
+      return -1;
     }
     /*
     ** clean up current pipe.
     */
-    pipes[cur].pmode = unopened;
+    *pipes[cur].pmode = '\0';
     free(pipes[cur].name);
     free(pipes[cur].command);
     return rval;
